/****************************************************************************
 ****************************************************************************
 *
 * ac.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <math.h>
#include <stdio.h>
#include <string.h>

#include "ac.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"



#define FLUSH_BITS			40
#define AC_SHIFT			13




/****************************************************************************
 *
 * misc helper routines
 *
 ****************************************************************************/




#ifdef AC_STATISTICS
/****************************************************************************
 * calc_size
 ****************************************************************************/
static inline double
calc_size(
	unsigned int			numerator,
	unsigned int			divisor)

	{
	return (-log(((double) numerator) / ((double) divisor)) / log(2));
	}



/****************************************************************************
 * percent
 ****************************************************************************/
static double
percent(
	int				count,
	int				sum)

	{
	if (sum == 0) return (0.0);
	return (100.0 * ((double) count) / ((double) sum));
	}
#endif /* AC_STATISTICS */




/****************************************************************************
 *
 * routines for arithmetic decoder and encoder
 *
 ****************************************************************************/




/****************************************************************************
 * ac_decode_init
 ****************************************************************************/
static int
ac_decode_init(
	struct arithmetic_codec		*arc,
	unsigned char			*data,
	int				size)

	{
	*arc        = (struct arithmetic_codec) { };
	arc->data   = data;
	arc->size   = size;
	arc->ofs    = 4;
	arc->length = 0xffffffff;
	arc->base   = arc->data[0];
	arc->base   = (arc->base << 8) | arc->data[1];
	arc->base   = (arc->base << 8) | arc->data[2];
	arc->base   = (arc->base << 8) | arc->data[3];
	return (0);
	}



/****************************************************************************
 * ac_decode
 ****************************************************************************/
static int
ac_decode(
	struct arithmetic_codec		*arc,
	int				prob0)

	{
	unsigned int			l = prob0 * (arc->length >> AC_SHIFT);
	int				bit = (arc->base >= l) ? 1 : 0;

	if (bit)
		{
		arc->base   -= l;
		arc->length -= l;
		}
	else arc->length = l;
	while (arc->length < 0x01000000)
		{
		if (arc->ofs >= arc->size) return (-1);
		arc->base   <<= 8;
		arc->length <<= 8;
		arc->base   |= arc->data[arc->ofs++];
		}
	return (bit);
	}



/****************************************************************************
 * ac_encode_init
 ****************************************************************************/
static int
ac_encode_init(
	struct arithmetic_codec		*arc,
	unsigned char			*data,
	int				size)

	{
	*arc = (struct arithmetic_codec) { .data = data, .size = size, .length = 0xffffffff };
	return (0);
	}



/****************************************************************************
 * ac_encode_carry
 ****************************************************************************/
static void
ac_encode_carry(
	struct arithmetic_codec		*arc)

	{
	int				i;

	for (i = arc->ofs - 1; arc->data[i] == 0xff; i--) arc->data[i] = 0;
	arc->data[i]++;
	}



/****************************************************************************
 * ac_encode
 ****************************************************************************/
static int
ac_encode(
	struct arithmetic_codec		*arc,
	int				prob0,
	int				bit)

	{
	unsigned int			b, l = prob0 * (arc->length >> AC_SHIFT);

	if (bit)
		{
		b           = arc->base;
		arc->base   += l;
		arc->length -= l;
		if (b > arc->base) ac_encode_carry(arc);
		}
	else arc->length = l;
	while (arc->length < 0x01000000)
		{
		if (arc->ofs >= arc->size) return (-1);
		arc->data[arc->ofs++] = arc->base >> 24;
		arc->base   <<= 8;
		arc->length <<= 8;
		}
	return (0);
	}



/****************************************************************************
 * ac_encode_flush
 ****************************************************************************/
static int
ac_encode_flush(
	struct arithmetic_codec		*arc)

	{
	int				i;

	for (i = 0; i < FLUSH_BITS; i++) ac_encode(arc, 1 << (AC_SHIFT - 1), 0);
	return (arc->ofs);
	}




/****************************************************************************
 *
 * routines for adaptivity
 *
 ****************************************************************************/




/****************************************************************************
 * ac_adaptive_reset
 ****************************************************************************/
static int
ac_adaptive_reset(
	struct adaptive_codec		*adc,
	int				bitcount,
	int				bitcount0,
	int				update)

	{
	debug_error_condition((bitcount < 2) || (bitcount0 < 1));
	debug_error_condition(bitcount > (1 << AC_SHIFT));
	debug_error_condition(bitcount0 >= bitcount);
	adc->bitcount          = bitcount;
	adc->bitcount0         = bitcount0;
	adc->bits_per_update   = update;
	adc->bits_until_update = update;
	return (0);
	}



/****************************************************************************
 * ac_adaptive_init
 ****************************************************************************/
static int
ac_adaptive_init(
	struct adaptive_codec		*adc,
	struct arithmetic_codec		*arc,
	int				update_interval,
	int				adaptive_shrink)

	{
	*adc = (struct adaptive_codec)
		{
		.arc             = arc,
		.prob0           = 1 << (AC_SHIFT - 1),
		.update_interval = update_interval,
		.adaptive_shrink = adaptive_shrink
		};
	return (ac_adaptive_reset(adc, 2, 1, 4));
	}



/****************************************************************************
 * ac_adaptive_update
 ****************************************************************************/
static int
ac_adaptive_update(
	struct adaptive_codec		*adc)

	{
	unsigned int			scale, bits, shrink = 0;

	adc->bitcount += adc->bits_per_update;
	if (adc->adaptive_shrink)
		{
		bits = adc->bitcount - adc->bitcount0;
		if (2 * bits >= adc->bitcount) bits = adc->bitcount - bits;
		if ((adc->bitcount >= 4000) && (adc->bitcount <= 400 * bits)) shrink = 1;
		if ((adc->bitcount >= 2000) && (adc->bitcount <= 200 * bits)) shrink = 1;
		if ((adc->bitcount >= 1000) && (adc->bitcount <= 100 * bits)) shrink = 1;
		if ((adc->bitcount >=  500) && (adc->bitcount <=  50 * bits)) shrink = 1;
		}
	if ((shrink) || (adc->bitcount >= 1 << AC_SHIFT))
		{
		adc->bitcount  = (adc->bitcount + 1) >> 1;
		adc->bitcount0 = (adc->bitcount0 + 1) >> 1;
		if (adc->bitcount0 == adc->bitcount) adc->bitcount++;
		}
	scale = 0x80000000 / adc->bitcount;
	adc->prob0 = (adc->bitcount0 * scale) >> (31 - AC_SHIFT);

	if (adc->update_interval == 0)
		{
		adc->bits_per_update = (5 * adc->bits_per_update) / 4;
		if (adc->bits_per_update > 64) adc->bits_per_update = 64;
		adc->bits_until_update = adc->bits_per_update;
		}
	else adc->bits_until_update = adc->bits_per_update = adc->update_interval;

	return (0);
	}



/****************************************************************************
 * ac_adaptive_decode
 ****************************************************************************/
static int
ac_adaptive_decode(
	struct adaptive_codec		*adc)

	{
	int				result = ac_decode(adc->arc, adc->prob0);

#ifdef AC_STATISTICS
	adc->bitcount_all++;
	if (result == 0)
		{
		adc->bitcount0_all++;
		adc->size += calc_size(adc->prob0, 1 << AC_SHIFT);
		}
	else adc->size += calc_size((1 << AC_SHIFT) - adc->prob0, 1 << AC_SHIFT);
#endif /* AC_STATISTICS */
	if (result == 0) adc->bitcount0++;
	if (--adc->bits_until_update == 0) ac_adaptive_update(adc);
	return (result);
	}



/****************************************************************************
 * ac_adaptive_encode
 ****************************************************************************/
static int
ac_adaptive_encode(
	struct adaptive_codec		*adc,
	int				bit)

	{
	int				result = ac_encode(adc->arc, adc->prob0, bit);

#ifdef AC_STATISTICS
	adc->bitcount_all++;
	if (bit == 0)
		{
		adc->bitcount0_all++;
		adc->size += calc_size(adc->prob0, 1 << AC_SHIFT);
		}
	else adc->size += calc_size((1 << AC_SHIFT) - adc->prob0, 1 << AC_SHIFT);
#endif /* AC_STATISTICS */
	if (bit == 0) adc->bitcount0++;
	if (--adc->bits_until_update == 0) ac_adaptive_update(adc);
	return (result);
	}




/****************************************************************************
 *
 * used by external callers
 *
 ****************************************************************************/




/****************************************************************************
 * ac_get_flush_bytes
 ****************************************************************************/
int
ac_get_flush_bytes(
	void)

	{
	return ((FLUSH_BITS + 7) / 8);
	}



/****************************************************************************
 * ac_init
 ****************************************************************************/
unsigned char *
ac_init(
	struct ac			*ac,
	unsigned char			*data,
	int				size,
	int				bitlimit,
	int				aclimit,
	int				mode,
	int				ctxs)

	{
	int				size2 = size;

	if (mode == AC_MODE_WRITE) size2 -= ac_get_flush_bytes();
	if (bitlimit <= 0) bitlimit = 0x7fffffff;
	if ((aclimit <= 0) || (aclimit >= size2)) aclimit = size2;
	*ac = (struct ac) { .bitlimit = bitlimit, .aclimit = aclimit, .mode = mode };
	verbose(1, "initializing bitstream with %d bytes for %d uncompressed bits and %d compressed bytes", size, bitlimit, aclimit);
	if (data == NULL)
		{
		data = (unsigned char *) malloc(size * sizeof (unsigned char));
		if (data == NULL) error_oom();
		memset(data, 0, size * sizeof (unsigned char));
		}
	if (mode == AC_MODE_READ) ac_decode_init(&ac->arc, data, size);
	else if (mode == AC_MODE_WRITE) ac_encode_init(&ac->arc, data, size);
	else debug_error();
	ac_init2(ac, ctxs, 0, 0);

	return (data);
	}



/****************************************************************************
 * ac_init2
 ****************************************************************************/
void
ac_init2(
	struct ac			*ac,
	int				ctxs,
	int				update_interval,
	int				adaptive_shrink)

	{
	int				ctx;

	debug_error_condition(ac->adc != NULL);
	if (ctxs <= 0) return;
	verbose(1, "initializing %d adaptive binary arithmetic contexts", ctxs);
	ac->adc  = (struct adaptive_codec *) malloc(ctxs * sizeof (struct adaptive_codec));
	ac->ctxs = ctxs;
	if (ac->adc == NULL) error_oom();
	for (ctx = 0; ctx < ctxs; ctx++) ac_adaptive_init(&ac->adc[ctx], &ac->arc, update_interval, adaptive_shrink);
	}



/****************************************************************************
 * ac_flush
 ****************************************************************************/
void
ac_flush(
	struct ac			*ac)

	{
	if (ac->mode == AC_MODE_WRITE) ac_encode_flush(&ac->arc);
	}



/****************************************************************************
 * ac_deinit
 ****************************************************************************/
void
ac_deinit(
	struct ac			*ac)

	{
#ifdef AC_STATISTICS
	struct adaptive_codec		*adc;
	int				ctx;
	double				size;

	for (ctx = 0; ctx < ac->ctxs; ctx++)
		{
		adc = &ac->adc[ctx];
		if (adc->bitcount_all == 0) continue;
		if (adc->bitcount0_all != 0)
			{
			int		count0 = adc->bitcount0_all;
			int		count  = adc->bitcount_all;
			
			size = ((double) count0) * calc_size(count0, count);
			size += ((double) (count - count0)) * calc_size(count - count0, count);
			}
		else size = 0.0;
		debug(1, "%4d: bitcount0_all =%8d, bitcount_all =%8d, %11.2f, size =%11.2f, %7.2f,%6.2f%%",
			ctx, adc->bitcount0_all, adc->bitcount_all, size, adc->size, adc->size - size,
			percent(adc->bitcount_all - adc->bitcount0_all, adc->bitcount_all));
		}
#endif /* AC_STATISTICS */
	if (ac->arc.data != NULL) free(ac->arc.data);
	if (ac->adc != NULL) free(ac->adc);
	}



/****************************************************************************
 * ac_read_bit
 ****************************************************************************/
int
ac_read_bit(
	struct ac			*ac)

	{
	debug_error_condition(ac->mode != AC_MODE_READ);
	if ((ac->bits >= ac->bitlimit) || (ac->arc.ofs >= ac->aclimit)) return (-1);
	ac->bits++;
	return (ac_decode(&ac->arc, 1 << (AC_SHIFT - 1)));
	}



/****************************************************************************
 * ac_write_bit
 ****************************************************************************/
int
ac_write_bit(
	struct ac			*ac,
	int				bit)

	{
	debug_error_condition(ac->mode != AC_MODE_WRITE);
	if ((ac->bits >= ac->bitlimit) || (ac->arc.ofs >= ac->aclimit)) return (-1);
	ac->bits++;
	return (ac_encode(&ac->arc, 1 << (AC_SHIFT - 1), bit));
	}



/****************************************************************************
 * ac_read_value
 ****************************************************************************/
int
ac_read_value(
	struct ac			*ac,
	int				len)

	{
	int				bit, val = 0;

	while (len-- > 0)
		{
		bit = ac_read_bit(ac);
		if (bit == -1) return (-1);
		val = (val << 1) | bit;
		}
	return (val);
	}



/****************************************************************************
 * ac_write_value
 ****************************************************************************/
int
ac_write_value(
	struct ac			*ac,
	int				val,
	int				len)

	{
	int				bit;

	while (len-- > 0)
		{
		bit = (val & (1 << len)) ? 1 : 0;
		if (ac_write_bit(ac, bit) == -1) return (-1);
		}
	return (0);
	}



/****************************************************************************
 * ac_ctx_reset
 ****************************************************************************/
int
ac_ctx_reset(
	struct ac			*ac,
	int				ctx,
	int				bitcount,
	int				bitcount0)

	{
	debug_error_condition(ctx >= ac->ctxs);
	return (ac_adaptive_reset(&ac->adc[ctx], bitcount, bitcount0, 4));
	}



/****************************************************************************
 * ac_ctx_read_bit
 ****************************************************************************/
int
ac_ctx_read_bit(
	struct ac			*ac,
	int				ctx)

	{
	debug_error_condition(ac->mode != AC_MODE_READ);
	debug_error_condition(ctx >= ac->ctxs);
	if ((ac->bits >= ac->bitlimit) || (ac->arc.ofs >= ac->aclimit)) return (-1);
	ac->bits++;
	return (ac_adaptive_decode(&ac->adc[ctx]));
	}



/****************************************************************************
 * ac_ctx_write_bit
 ****************************************************************************/
int
ac_ctx_write_bit(
	struct ac			*ac,
	int				ctx,
	int				bit)

	{
	debug_error_condition(ac->mode != AC_MODE_WRITE);
	debug_error_condition(ctx >= ac->ctxs);
	if ((ac->bits >= ac->bitlimit) || (ac->arc.ofs >= ac->aclimit)) return (-1);
	ac->bits++;
	return (ac_adaptive_encode(&ac->adc[ctx], bit));
	}



/****************************************************************************
 * ac_get_data
 ****************************************************************************/
unsigned char *
ac_get_data(
	struct ac			*ac)

	{
	return (ac->arc.data);
	}



/****************************************************************************
 * ac_get_uncompressed_bits
 ****************************************************************************/
int
ac_get_uncompressed_bits(
	struct ac			*ac)

	{
	return (ac->bits);
	}



/****************************************************************************
 * ac_get_compressed_bytes
 ****************************************************************************/
int
ac_get_compressed_bytes(
	struct ac			*ac)

	{
	return (ac->arc.ofs);
	}
/******************************************************** Karsten Scheibler */
