/****************************************************************************
 ****************************************************************************
 *
 * codec002.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>

#include "codec002.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "codec.h"
#include "picture.h"



#define CTX_BLOCK			0
#define CTX_SIGNIF			1
#define CTX_SIGN			2
#define CTX_MAGN			3

struct block
	{
	int				xstart;
	int				ystart;
	int				xend;
	int				yend;
	int				msb;
	};



/****************************************************************************
 * block_too_small
 ****************************************************************************/
static int
block_too_small(
	int				xstart,
	int				ystart,
	int				xend,
	int				yend)

	{
	if ((xend - xstart) * (yend - ystart) < 256) return (1);
	return (0);
	}



/****************************************************************************
 * block_get_significance
 ****************************************************************************/
static int
block_get_significance(
	struct block			*blk,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				bitplane)

	{
	int				i;

	for (i = 0; blk[i].xend + blk[i].yend > 0; i++)
		{
		if ((xstart == blk[i].xstart) && (ystart == blk[i].ystart) &&
			(xend == blk[i].xend) && (yend == blk[i].yend)) break;
		}
	debug_error_condition(blk[i].msb > bitplane);
	if (blk[i].msb == bitplane) return (1);
	return (0);
	}



/****************************************************************************
 * block_already_significant
 ****************************************************************************/
static int
block_already_significant(
	struct block			*blk,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend)

	{
	int				i;

	for (i = 0; blk[i].xend + blk[i].yend > 0; i++)
		{
		if ((xstart == blk[i].xstart) && (ystart == blk[i].ystart) &&
			(xend == blk[i].xend) && (yend == blk[i].yend)) return (1);
		}
	return (0);
	}



/****************************************************************************
 * block_append
 ****************************************************************************/
static void
block_append(
	struct block			*blk,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend)

	{
	int				i;

	for (i = 0; blk[i].xend + blk[i].yend > 0; i++) ;
	blk[i] = (struct block) { .xstart = xstart, .ystart = ystart, .xend = xend, .yend = yend };
	}



/****************************************************************************
 * block_msb
 ****************************************************************************/
static void
block_msb(
	struct codec			*cdc,
	struct block			*blk,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend)

	{
	int				i, x, y, m;

	*blk = (struct block) { .xstart = xstart, .ystart = ystart, .xend = xend, .yend = yend };
	for (y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		m = pcoeff_get_msb(&cdc->cft[i].pcf);
		if (m > blk->msb) blk->msb = m;
		}
	debug(1, "(%d,%d)-(%d,%d) msb = %d", xstart, ystart, xend, yend, blk->msb);
	}



/****************************************************************************
 * block_encode
 ****************************************************************************/
static int
block_encode(
	struct codec			*cdc,
	struct block			*blk,
	struct block			*blk2,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				bitplane)

	{
	struct pcoeff			*pcf;
	int				bit, i, j, x, y;

	if (block_too_small(xstart, ystart, xend, yend)) goto skip;
	if (block_already_significant(blk2, xstart, ystart, xend, yend)) goto skip;
	bit = block_get_significance(blk, xstart, ystart, xend, yend, bitplane);
	if (ac_ctx_write_bit(cdc->ac, CTX_BLOCK, bit) == -1) return (-1);
	if (! bit) return (0);
	block_append(blk2, xstart, ystart, xend, yend);
skip:
	for (j = 0, y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcf = &cdc->cft[i].pcf;
		if (pcoeff_already_finished(pcf, bitplane)) continue;
		j++;
		if (pcoeff_already_significant(pcf, bitplane))
			{
			bit = pcoeff_get_magnitude(pcf, bitplane);
			if (ac_ctx_write_bit(cdc->ac, CTX_MAGN, bit) == -1) return (-1);
			cdc->magn++;
			continue;
			}
		bit = pcoeff_get_significance(pcf, bitplane);
		if (ac_ctx_write_bit(cdc->ac, CTX_SIGNIF, bit) == -1) return (-1);
		cdc->signif++;
		if (! bit) continue;
		bit = pcoeff_get_sign(pcf);
		if (ac_ctx_write_bit(cdc->ac, CTX_SIGN, bit) == -1) return (-1);
		cdc->sign++;
		}
	debug(1, "(%d,%d)-(%d,%d) j = %d", xstart, ystart, xend, yend, j);
	return (0);
	}



/****************************************************************************
 * block_decode
 ****************************************************************************/
static int
block_decode(
	struct codec			*cdc,
	struct block			*blk,
	struct block			*blk2,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				bitplane)

	{
	struct pcoeff			*pcf;
	int				bit, i, j, x, y;

	if (block_too_small(xstart, ystart, xend, yend)) goto skip;
	if (block_already_significant(blk2, xstart, ystart, xend, yend)) goto skip;
	bit = ac_ctx_read_bit(cdc->ac, CTX_BLOCK);
	if (bit == -1) return (-1);
	if (! bit) return (0);
	block_append(blk2, xstart, ystart, xend, yend);
skip:
	for (j = 0, y = ystart; y < yend; y++) for (i = y * cdc->xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcf = &cdc->cft[i].pcf;
		if (pcoeff_already_finished(pcf, bitplane)) continue;
		j++;
		if (pcoeff_already_significant(pcf, bitplane))
			{
			bit = ac_ctx_read_bit(cdc->ac, CTX_MAGN);
			if (bit == -1) return (-1);
			pcoeff_set_magnitude(pcf, bitplane, bit);
			cdc->magn++;
			continue;
			}
		bit = ac_ctx_read_bit(cdc->ac, CTX_SIGNIF);
		if (bit == -1) return (-1);
		cdc->signif++;
		if (! bit) continue;
		bit = ac_ctx_read_bit(cdc->ac, CTX_SIGN);
		if (bit == -1) return (-1);
		pcoeff_set_significance_and_sign(pcf, bitplane, bit);
		cdc->sign++;
		}
	debug(1, "(%d,%d)-(%d,%d) j = %d", xstart, ystart, xend, yend, j);
	return (0);
	}



/****************************************************************************
 * process_bitplane
 ****************************************************************************/
static int
process_bitplane(
	struct codec			*cdc,
	struct block			*blk,
	struct block			*blk2,
	int				(*func)(struct codec *, struct block *, struct block *, int, int, int, int, int),
	int				bitplane)

	{
	int				x = cdc->xres >> cdc->level;
	int				y = cdc->yres >> cdc->level;
	int				l;

	if (func(cdc, blk, blk2, 0, 0, x, y, bitplane) == -1) return (-1);
	for (l = 0; l < cdc->level; l++, x <<= 1, y <<= 1)
		{
		if (func(cdc, blk, blk2, x, 0, x + x,     y, bitplane) == -1) return (-1);
		if (func(cdc, blk, blk2, 0, y,     x, y + y, bitplane) == -1) return (-1);
		if (func(cdc, blk, blk2, x, y, x + x, y + y, bitplane) == -1) return (-1);
		}
	return (0);
	}



/****************************************************************************
 * iterate
 ****************************************************************************/
static void
iterate(
	struct codec			*cdc,
	struct block			*blk,
	int				(*func)(struct codec *, struct block *, struct block *, int, int, int, int, int))

	{
	struct block			blk2[64] = { };
	int				b, n, result;

	ac_init2(cdc->ac, 4, 0, 0);
	for (b = cdc->msb, n = 0; (b >= cdc->bitplane) && (n < cdc->nr_bitplanes); b--, n++)
		{
		cdc->signif = cdc->sign = cdc->magn = 0;
		result = process_bitplane(cdc, blk, blk2, func, b);
		cdc->sum_signif += cdc->signif;
		cdc->sum_sign   += cdc->sign;
		cdc->sum_magn   += cdc->magn;
		verbose(2, "bitplane %2d: signif =%9d/%-9d, sign =%9d/%-9d, magn =%9d/%-9d",
			b, cdc->signif, cdc->sum_signif, cdc->sign, cdc->sum_sign, cdc->magn, cdc->sum_magn);
		if (result == -1) break;
		}
	}



/****************************************************************************
 * codec002_encode
 ****************************************************************************/
void
codec002_encode(
	struct codec			*cdc)

	{
	struct block			blk[64] = { };
	int				x = cdc->xres;
	int				y = cdc->yres;
	int				i, l;

	debug_error_condition(cdc->level > 16);
	for (i = l = 0; l < cdc->level; l++)
		{
		x >>= 1;
		y >>= 1;
		block_msb(cdc, &blk[i++], x, 0, x + x,     y);
		block_msb(cdc, &blk[i++], 0, y,     x, y + y);
		block_msb(cdc, &blk[i++], x, y, x + x, y + y);
		if (block_too_small(0, 0, x >> 1, y >> 1)) break;
		}
	block_msb(cdc, &blk[i], 0, 0, x, y);
	for (cdc->msb = 0; blk[i].xend + blk[i].yend > 0; i++) if (blk[i].msb > cdc->msb) cdc->msb = blk[i].msb;
	iterate(cdc, blk, block_encode);
	}



/****************************************************************************
 * codec002_decode
 ****************************************************************************/
void
codec002_decode(
	struct codec			*cdc)

	{
	debug_error_condition(cdc->level > 16);
	iterate(cdc, NULL, block_decode);
	}
/******************************************************** Karsten Scheibler */
