/****************************************************************************
 ****************************************************************************
 *
 * file.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#include "file.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "codec.h"
#include "picture.h"
#include "transform.h"



struct header
	{
	char				magic[7];
	unsigned char			bitlimit[4];
	unsigned char			xres[2];
	unsigned char			yres[2];
	unsigned char			level;
	unsigned char			codec;
	unsigned char			msb;
	unsigned char			shift;
	unsigned char			wavelet;
	};

static const char			file_magic[] = "yawica";



/****************************************************************************
 * write_info1
 ****************************************************************************/
static void
write_info1(
	char				*action,
	char				*path,
	struct file			*fil)

	{
	verbose(1, "%s file '%s', codec: %d, wavelet: %d, level(s): %d, resolution: %dx%d, bitlimit: %d, shift: %d, msb: %d",
		action, path, fil->codec, fil->wavelet, fil->level,
		fil->xres, fil->yres, fil->bitlimit, fil->shift, fil->msb);
	}



/****************************************************************************
 * write_info2
 ****************************************************************************/
static void
write_info2(
	char				*action,
	char				*path,
	int				len)

	{
	verbose(1, "%s file '%s', bytes: %d", action, path, len);
	}



/****************************************************************************
 * file_get_header_size
 ****************************************************************************/
int
file_get_header_size(
	void)

	{
	return (sizeof (struct header));
	}



/****************************************************************************
 * file_init
 ****************************************************************************/
void
file_init(
	struct file			*fil,
	struct transform		*trf,
	struct codec			*cdc)

	{
	struct picture			*pic = transform_get_picture(trf);
	struct ac			*ac = codec_get_ac(cdc);

	*fil = (struct file)
		{
		.data     = ac_get_data(ac),
		.size     = ac_get_compressed_bytes(ac),
		.bitlimit = ac_get_uncompressed_bits(ac),
		.xres     = picture_get_xres(pic),
		.yres     = picture_get_yres(pic),
		.level    = transform_get_level(trf),
		.codec    = codec_get_number(cdc),
		.msb      = codec_get_msb(cdc),
		.shift    = picture_get_shift(pic),
		.wavelet  = transform_get_wavelet(trf)
		};
	}



/****************************************************************************
 * file_read
 ****************************************************************************/
void
file_read(
	struct file			*fil,
	char				*path)

	{
	FILE				*fh;
	struct header			h;
	struct stat			s;

	fh = fopen(path, "r");
	if (fh == NULL) error_perror2("error while opening '%s'", path);

	/* read header */

	fread(&h, sizeof (h), 1, fh);
	if (strncmp(h.magic, file_magic, sizeof (h.magic)) != 0) error("'%s' is not a valid compressed file", path);
	*fil = (struct file)
		{
		.bitlimit = 0x1000000 * ((int) h.bitlimit[3]) + 0x10000 * ((int) h.bitlimit[2]) + 0x100 * ((int) h.bitlimit[1]) + ((int) h.bitlimit[0]),
		.xres     = 0x100 * ((int) h.xres[1]) + ((int) h.xres[0]),
		.yres     = 0x100 * ((int) h.yres[1]) + ((int) h.yres[0]),
		.level    = h.level,
		.codec    = h.codec,
		.msb      = h.msb,
		.shift    = h.shift,
		.wavelet  = h.wavelet
		};
	write_info1("reading", path, fil);

	/* get real file size (file may be truncated) */

	if (stat(path, &s) == -1) error_perror2("error while accessing '%s'", path);
	fil->size = (s.st_size - sizeof (h)) * sizeof (unsigned char);

	/* allocate memory and read data */

	fil->data = (unsigned char *) malloc(fil->size * sizeof (unsigned char));
	if (fil->data == NULL) error_oom();
	fread(fil->data, fil->size, 1, fh);
	fclose(fh);
	write_info2("read", path, fil->size + sizeof (h));
	}



/****************************************************************************
 * file_write
 ****************************************************************************/
void
file_write(
	struct file			*fil,
	char				*path)

	{
	struct header			h;
	FILE				*fh;

	debug_error_condition(fil->xres > 0xffff);
	debug_error_condition(fil->yres > 0xffff);
	debug_error_condition(fil->level > 0xff);
	debug_error_condition(fil->codec > 0xff);
	debug_error_condition(fil->msb > 0xff);
	debug_error_condition(fil->shift > 0xff);
	debug_error_condition(fil->wavelet > 0xff);
	write_info1("writing", path, fil);
	h = (struct header)
		{
		.bitlimit = { fil->bitlimit % 0x100, (fil->bitlimit / 0x100) % 0x100, (fil->bitlimit / 0x10000) % 0x100, (fil->bitlimit / 0x1000000) % 0x100 },
		.xres     = { fil->xres % 0x100, fil->xres / 0x100 },
		.yres     = { fil->yres % 0x100, fil->yres / 0x100 },
		.level    = fil->level,
		.codec    = fil->codec,
		.msb      = fil->msb,
		.shift    = fil->shift,
		.wavelet  = fil->wavelet
		};
	strncpy(h.magic, file_magic, sizeof (h.magic));
	fh = fopen(path, "w");
	if (fh == NULL) error_perror2("error while opening '%s'", path);
	fwrite(&h, sizeof (h), 1, fh);
	fwrite(fil->data, fil->size, 1, fh);
	fclose(fh);
	write_info2("wrote", path, fil->size + sizeof (h));
	}



/****************************************************************************
 * file_get_data
 ****************************************************************************/
unsigned char *
file_get_data(
	struct file			*fil)

	{
	return (fil->data);
	}



/****************************************************************************
 * file_get_size
 ****************************************************************************/
int
file_get_size(
	struct file			*fil)

	{
	return (fil->size);
	}



/****************************************************************************
 * file_get_bitlimit
 ****************************************************************************/
int
file_get_bitlimit(
	struct file			*fil)

	{
	return (fil->bitlimit);
	}



/****************************************************************************
 * file_get_xres
 ****************************************************************************/
int
file_get_xres(
	struct file			*fil)

	{
	return (fil->xres);
	}



/****************************************************************************
 * file_get_yres
 ****************************************************************************/
int
file_get_yres(
	struct file			*fil)

	{
	return (fil->yres);
	}



/****************************************************************************
 * file_get_level
 ****************************************************************************/
int
file_get_level(
	struct file			*fil)

	{
	return (fil->level);
	}



/****************************************************************************
 * file_get_codec
 ****************************************************************************/
int
file_get_codec(
	struct file			*fil)

	{
	return (fil->codec);
	}



/****************************************************************************
 * file_get_msb
 ****************************************************************************/
int
file_get_msb(
	struct file			*fil)

	{
	return (fil->msb);
	}



/****************************************************************************
 * file_get_shift
 ****************************************************************************/
int
file_get_shift(
	struct file			*fil)

	{
	return (fil->shift);
	}



/****************************************************************************
 * file_get_wavelet
 ****************************************************************************/
int
file_get_wavelet(
	struct file			*fil)

	{
	return (fil->wavelet);
	}
/******************************************************** Karsten Scheibler */
