/****************************************************************************
 ****************************************************************************
 *
 * transform.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "transform.h"
#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "picture.h"




/****************************************************************************
 *
 * helper routines for transformation
 *
 ****************************************************************************/




#define ALPHA				-1.58615986717275
#define BETA				-0.05297864003258
#define GAMMA				0.88293362717904
#define DELTA				0.44350482244527
#define EPSILON				1.14960430535816



/****************************************************************************
 * do_round
 ****************************************************************************/
static int
do_round(
	int				val1,
	int				val2)

	{
	if (val1 < 0) return (val1 - val2);
	return (val1 + val2);
	}



/****************************************************************************
 * transform_forward_haar
 ****************************************************************************/
static void
transform_forward_haar(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	int				*low, *high;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (int *) alloca(m * sizeof (int));
	high = (int *) alloca(m * sizeof (int));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = j = 0; i < m; i++, j += 2)
		{
		low[i]  = (int) cft[j * ofs].d;
		high[i] = (int) cft[(j + 1) * ofs].d;
		}

	for (i = 0; i < m; i++)
		{
		high[i] -= low[i];
		low[i] += do_round(high[i], 1) / 2;
		}

	for (i = 0; i < m; i++)
		{
		cft[i * ofs].d = (double) low[i];
		cft[(m + i) * ofs].d = (double) high[i];
		}
	}



/****************************************************************************
 * transform_reverse_haar
 ****************************************************************************/
static void
transform_reverse_haar(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	int				*low, *high;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (int *) alloca(m * sizeof (int));
	high = (int *) alloca(m * sizeof (int));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = 0; i < m; i++)
		{
		low[i]  = (int) cft[i * ofs].d;
		high[i] = (int) cft[(m + i) * ofs].d;
		}

	for (i = m; i-- > 0; )
		{
		low[i] -= do_round(high[i], 1) / 2;
		high[i] += low[i];
		}

	for (i = j = 0; i < m; i++, j += 2)
		{
		cft[j * ofs].d = (double) low[i];
		cft[(j + 1) * ofs].d = (double) high[i];
		}
	}



/****************************************************************************
 * transform_forward_sp
 ****************************************************************************/
static void
transform_forward_sp(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	int				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (int *) alloca(m * sizeof (int));
	high = (int *) alloca(m * sizeof (int));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = j = 0; i < m; i++, j += 2)
		{
		low[i]  = (int) cft[j * ofs].d;
		high[i] = (int) cft[(j + 1) * ofs].d;
		}

	for (i = 0; i < m; i++)
		{
		high[i] -= low[i];
		low[i] += do_round(high[i], 1) / 2;
		}

	if (m > 1) for (i = 0; i < m; i++)
		{
		if (i == 0) val = do_round(low[0] - low[1], 2) / 4;
		else if (i + 1 == m) val = do_round(low[i - 1] - low[i], 2) / 4;
		else val = do_round(2 * low[i - 1] + low[i] - 3 * low[i + 1] + 2 * high[i + 1], 4) / 8;
		high[i] += val;
		}

	for (i = 0; i < m; i++)
		{
		cft[i * ofs].d = (double) low[i];
		cft[(m + i) * ofs].d = (double) high[i];
		}
	}



/****************************************************************************
 * transform_reverse_sp
 ****************************************************************************/
static void
transform_reverse_sp(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	int				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (int *) alloca(m * sizeof (int));
	high = (int *) alloca(m * sizeof (int));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = 0; i < m; i++)
		{
		low[i]  = (int) cft[i * ofs].d;
		high[i] = (int) cft[(m + i) * ofs].d;
		}

	if (m > 1) for (i = m; i-- > 0; )
		{
		if (i == 0) val = do_round(low[0] - low[1], 2) / 4;
		else if (i + 1 == m) val = do_round(low[i - 1] - low[i], 2) / 4;
		else val = do_round(2 * low[i - 1] + low[i] - 3 * low[i + 1] + 2 * high[i + 1], 4) / 8;
		high[i] -= val;
		}

	for (i = m; i-- > 0; )
		{
		low[i] -= do_round(high[i], 1) / 2;
		high[i] += low[i];
		}

	for (i = j = 0; i < m; i++, j += 2)
		{
		cft[j * ofs].d = (double) low[i];
		cft[(j + 1) * ofs].d = (double) high[i];
		}
	}



/****************************************************************************
 * transform_forward_53
 ****************************************************************************/
static void
transform_forward_53(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	int				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (int *) alloca(m * sizeof (int));
	high = (int *) alloca(m * sizeof (int));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = j = 0; i < m; i++, j += 2)
		{
		low[i]  = (int) cft[j * ofs].d;
		high[i] = (int) cft[(j + 1) * ofs].d;
		}

	for (i = 0; i < m; i++)
		{
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] -= do_round(low[i] + val, 1) / 2;
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] += do_round(val + high[i], 2) / 4;
		}

	for (i = 0; i < m; i++)
		{
		cft[i * ofs].d = (double) low[i];
		cft[(m + i) * ofs].d = (double) high[i];
		}
	}



/****************************************************************************
 * transform_reverse_53
 ****************************************************************************/
static void
transform_reverse_53(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	int				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (int *) alloca(m * sizeof (int));
	high = (int *) alloca(m * sizeof (int));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = 0; i < m; i++)
		{
		low[i]  = (int) cft[i * ofs].d;
		high[i] = (int) cft[(m + i) * ofs].d;
		}

	for (i = m; i-- > 0; )
		{
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] -= do_round(val + high[i], 2) / 4;
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] += do_round(low[i] + val, 1) / 2;
		}

	for (i = j = 0; i < m; i++, j += 2)
		{
		cft[j * ofs].d = (double) low[i];
		cft[(j + 1) * ofs].d = (double) high[i];
		}
	}



/****************************************************************************
 * transform_forward_53l
 ****************************************************************************/
static void
transform_forward_53l(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	double				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (double *) alloca(m * sizeof (double));
	high = (double *) alloca(m * sizeof (double));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = j = 0; i < m; i++, j += 2)
		{
		low[i]  = cft[j * ofs].d;
		high[i] = cft[(j + 1) * ofs].d;
		}

	for (i = 0; i < m; i++)
		{
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] -= (low[i] + val) / 2;
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] += (val + high[i]) / 4;
		}

	for (i = 0; i < m; i++)
		{
		cft[i * ofs].d = M_SQRT2 * low[i];
		cft[(m + i) * ofs].d = high[i] / -M_SQRT2;
		}
	}



/****************************************************************************
 * transform_reverse_53l
 ****************************************************************************/
static void
transform_reverse_53l(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	double				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (double *) alloca(m * sizeof (double));
	high = (double *) alloca(m * sizeof (double));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = 0; i < m; i++)
		{
		low[i]  = cft[i * ofs].d / M_SQRT2;
		high[i] = -M_SQRT2 * cft[(m + i) * ofs].d;
		}

	for (i = m; i-- > 0; )
		{
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] -= (val + high[i]) / 4;
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] += (low[i] + val) / 2;
		}

	for (i = j = 0; i < m; i++, j += 2)
		{
		cft[j * ofs].d = low[i];
		cft[(j + 1) * ofs].d = high[i];
		}
	}



/****************************************************************************
 * transform_forward_97
 ****************************************************************************/
static void
transform_forward_97(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	double				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (double *) alloca(m * sizeof (double));
	high = (double *) alloca(m * sizeof (double));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = j = 0; i < m; i++, j += 2)
		{
		low[i]  = cft[j * ofs].d;
		high[i] = cft[(j + 1) * ofs].d;
		}

	for (i = 0; i < m; i++)
		{
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] += ALPHA * (low[i] + val);
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] += BETA * (high[i] + val);
		}

	for (i = 0; i < m; i++)
		{
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] += GAMMA * (low[i] + val);
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] += DELTA * (high[i] + val);
		}

	for (i = 0; i < m; i++)
		{
		cft[i * ofs].d = EPSILON * low[i];
		cft[(m + i) * ofs].d = high[i] / -EPSILON;
		}
	}



/****************************************************************************
 * transform_reverse_97
 ****************************************************************************/
static void
transform_reverse_97(
	union coefficient		*cft,
	int				ofs,
	int				len)

	{
	int				i, j, m;
	double				*low, *high, val;

	debug_error_condition((len % 2) != 0);
	m = len / 2;
	low  = (double *) alloca(m * sizeof (double));
	high = (double *) alloca(m * sizeof (double));
	if ((low == NULL) || (high == NULL)) error_oom();

	for (i = 0; i < m; i++)
		{
		low[i]  = cft[i * ofs].d / EPSILON;
		high[i] = -EPSILON * cft[(m + i) * ofs].d;
		}

	for (i = m; i-- > 0; )
		{
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] -= DELTA * (high[i] + val);
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] -= GAMMA * (low[i] + val);
		}

	for (i = m; i-- > 0; )
		{
		val = (i > 0) ? high[i - 1] : high[i];
		low[i] -= BETA * (high[i] + val);
		val = (i + 1 < m) ? low[i + 1] : low[i];
		high[i] -= ALPHA * (low[i] + val);
		}

	for (i = j = 0; i < m; i++, j += 2)
		{
		cft[j * ofs].d = low[i];
		cft[(j + 1) * ofs].d = high[i];
		}
	}



/****************************************************************************
 * write_info
 ****************************************************************************/
static void
write_info(
	char				*direction,
	int				wavelet,
	int				level,
	int				xres,
	int				yres)

	{
	char				*name = "Haar";

	if (wavelet == TRANSFORM_SP)          name = "S+P";
	if (wavelet == TRANSFORM_WAVELET_53)  name = "5/3";
	if (wavelet == TRANSFORM_WAVELET_53L) name = "5/3l";
	if (wavelet == TRANSFORM_WAVELET_97)  name = "9/7";
	verbose(1, "transforming %s, wavelet: %d (%s), level(s): %d, resolution: %dx%d", direction, wavelet, name, level, xres, yres);
	}




/****************************************************************************
 *
 * helper routines for conversion
 *
 ****************************************************************************/




/****************************************************************************
 * convert_block_weight
 ****************************************************************************/
static void
convert_block_weight(
	union coefficient		*cft,
	int				xres,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				lsb)

	{
	int				i, x, y;

	debug(1, "(%d,%d)-(%d,%d) lsb = %d", xstart, ystart, xend, yend, lsb);
	for (y = ystart; y < yend; y++) for (i = y * xres + xstart, x = xstart; x < xend; i++, x++)
		{
		pcoeff_set(&cft[i].pcf, cft[i].d, lsb);
		}
	}



/****************************************************************************
 * unconvert_block_weight
 ****************************************************************************/
static void
unconvert_block_weight(
	union coefficient		*cft,
	int				xres,
	int				xstart,
	int				ystart,
	int				xend,
	int				yend,
	int				lsb)

	{
	int				i, x, y;

	debug(1, "(%d,%d)-(%d,%d) lsb = %d", xstart, ystart, xend, yend, lsb);
	for (y = ystart; y < yend; y++) for (i = y * xres + xstart, x = xstart; x < xend; i++, x++)
		{
		cft[i].d = pcoeff_get(&cft[i].pcf);
		}
	}



/****************************************************************************
 * block_weight
 ****************************************************************************/
static void
block_weight(
	struct transform		*trf,
	union coefficient		*cft,
	void				(*func)(union coefficient *, int, int, int, int, int, int))

	{
	int				xres = picture_get_xres(trf->pic);
	int				l    = trf->level;
	int				x    = picture_get_xres(trf->pic) >> l;
	int				y    = picture_get_yres(trf->pic) >> l;

	func(cft, xres, 0, 0, x, y, l + 1);
	for ( ; l-- > 0; x <<= 1, y <<= 1)
		{
		func(cft, xres, x, 0, x + x,     y, l + 1);
		func(cft, xres, 0, y,     x, y + y, l + 1);
		func(cft, xres, x, y, x + x, y + y, l);
		}
	}




/****************************************************************************
 *
 * used by external callers
 *
 ****************************************************************************/




/****************************************************************************
 * transform_get_max_level
 ****************************************************************************/
int
transform_get_max_level(
	int				xres,
	int				yres,
	int				level)

	{
	int				i, j;

	for (i = 0; (xres % 2) == 0; i++) xres >>= 1;
	for (j = 0; (yres % 2) == 0; j++) yres >>= 1;
	if (i > j) i = j;
	if (level < 1) level = i;
	if (level > i) error_message("level too large for picture resolution");
	return (level);
	}



/****************************************************************************
 * transform_get_wavelet_number
 ****************************************************************************/
int
transform_get_wavelet_number(
	char				*name)

	{
	if (name == NULL) return (TRANSFORM_NONE);
	if (strcasecmp(name, "haar") == 0) return (TRANSFORM_WAVELET_HAAR);
	if (strcasecmp(name, "s+p")  == 0) return (TRANSFORM_SP);
	if (strcasecmp(name, "5/3")  == 0) return (TRANSFORM_WAVELET_53);
	if (strcasecmp(name, "5/3l") == 0) return (TRANSFORM_WAVELET_53L);
	if (strcasecmp(name, "9/7")  == 0) return (TRANSFORM_WAVELET_97);
	return (TRANSFORM_NONE);
	}



/****************************************************************************
 * transform_init
 ****************************************************************************/
void
transform_init(
	struct transform		*trf,
	struct picture			*pic,
	int				level,
	int				wavelet)

	{
	transform_get_max_level(picture_get_xres(pic), picture_get_yres(pic), level);
	if (level == 0) error_message("invalid level");
	if ((wavelet != TRANSFORM_WAVELET_HAAR) && (wavelet != TRANSFORM_SP) &&
		(wavelet != TRANSFORM_WAVELET_53) && (wavelet != TRANSFORM_WAVELET_53L) &&
		(wavelet != TRANSFORM_WAVELET_97)) error_message("invalid wavelet");
	*trf = (struct transform) { .pic = pic, .level = level, .wavelet = wavelet };
	}



/****************************************************************************
 * transform_forward
 ****************************************************************************/
void
transform_forward(
	struct transform		*trf)

	{
	union coefficient		*cft;
	int				x, y, xlen, ylen, xres, yres;
	int				level, wavelet;
	void				(*func)(union coefficient *, int, int) = transform_forward_haar;

	cft     = picture_get_coefficient(trf->pic, PICTURE_MODE_DOUBLE);
	xres    = picture_get_xres(trf->pic);
	yres    = picture_get_yres(trf->pic);
	level   = trf->level;
	wavelet = trf->wavelet;
	write_info("forward", wavelet, level, xres, yres);
	if (wavelet == TRANSFORM_SP)          func = transform_forward_sp;
	if (wavelet == TRANSFORM_WAVELET_53)  func = transform_forward_53;
	if (wavelet == TRANSFORM_WAVELET_53L) func = transform_forward_53l;
	if (wavelet == TRANSFORM_WAVELET_97)  func = transform_forward_97;
	for (xlen = xres, ylen = yres; level > 0; level--)
		{
		debug(1, "(0,0)-(%d,%d)", xlen, ylen);
		for (y = 0; y < ylen; y++) func(&cft[y * xres],    1, xlen);
		for (x = 0; x < xlen; x++) func(&cft[x],        xres, ylen);
		xlen >>= 1;
		ylen >>= 1;
		}
	}



/****************************************************************************
 * transform_reverse
 ****************************************************************************/
void
transform_reverse(
	struct transform		*trf)

	{
	union coefficient		*cft;
	int				x, y, xlen, ylen, xres, yres;
	int				level, wavelet;
	void				(*func)(union coefficient *, int, int) = transform_reverse_haar;

	cft     = picture_get_coefficient(trf->pic, PICTURE_MODE_DOUBLE);
	xres    = picture_get_xres(trf->pic);
	yres    = picture_get_yres(trf->pic);
	level   = trf->level;
	wavelet = trf->wavelet;
	write_info("reverse", wavelet, level, xres, yres);
	if (wavelet == TRANSFORM_SP)          func = transform_reverse_sp;
	if (wavelet == TRANSFORM_WAVELET_53)  func = transform_reverse_53;
	if (wavelet == TRANSFORM_WAVELET_53L) func = transform_reverse_53l;
	if (wavelet == TRANSFORM_WAVELET_97)  func = transform_reverse_97;
	for (xlen = xres >> level, ylen = yres >> level; level > 0; level--)
		{
		xlen <<= 1;
		ylen <<= 1;
		debug(1, "(0,0)-(%d,%d)", xlen, ylen);
		for (x = 0; x < xlen; x++) func(&cft[x],        xres, ylen);
		for (y = 0; y < ylen; y++) func(&cft[y * xres],    1, xlen);
		}
	}



/****************************************************************************
 * transform_convert
 ****************************************************************************/
void
transform_convert(
	struct transform		*trf)

	{
	union coefficient		*cft = picture_get_coefficient(trf->pic, PICTURE_MODE_DOUBLE);
	int				xres = picture_get_xres(trf->pic);
	int				yres = picture_get_yres(trf->pic);
	int				i, j = xres * yres;
	double				m = 0.0, val;

	verbose(1, "converting from double to pcoeff, resolution %dx%d", xres, yres);
	if ((trf->wavelet != TRANSFORM_WAVELET_53L) && (trf->wavelet != TRANSFORM_WAVELET_97))
		{
		block_weight(trf, cft, convert_block_weight);
		goto end;
		}

	/* get largest absolute value */

	for (i = 0; i < j; i++)
		{
		val = fabs(cft[i].d);
		if (val > m) m = val;
		}
	debug(1, "largest absolute value is %1.9f", m);

	/* determine multiplicator */

	for (i = (int) m, j = -1; i > 0; j++) i >>= 1;
	if (j == -1) j = 0;
	else j = 31 - j;
	picture_set_shift(trf->pic, j);
	m = (double) (1 << j);
	debug(1, "multiplier is %1.9f", m);

	/* convert */

	for (i = 0, j = xres * yres; i < j; i++) pcoeff_set(&cft[i].pcf, cft[i].d * m, 0);
end:
	picture_set_mode(trf->pic, PICTURE_MODE_PCOEFF);
	}



/****************************************************************************
 * transform_unconvert
 ****************************************************************************/
void
transform_unconvert(
	struct transform		*trf)

	{
	union coefficient		*cft = picture_get_coefficient(trf->pic, PICTURE_MODE_PCOEFF);
	int				xres = picture_get_xres(trf->pic);
	int				yres = picture_get_yres(trf->pic);
	int				i, j = xres * yres;
	double				d;

	verbose(1, "converting from pcoeff to double, resolution %dx%d", xres, yres);
	if ((trf->wavelet != TRANSFORM_WAVELET_53L) && (trf->wavelet != TRANSFORM_WAVELET_97))
		{
		block_weight(trf, cft, unconvert_block_weight);
		goto end;
		}

	/* determine divisor and unconvert */

	d = (double) (1 << picture_get_shift(trf->pic));
	debug(1, "divisor is %1.9f", d);
	for (i = 0; i < j; i++) cft[i].d = pcoeff_get(&cft[i].pcf) / d;
end:
	picture_set_mode(trf->pic, PICTURE_MODE_DOUBLE);
	}



/****************************************************************************
 * transform_get_picture
 ****************************************************************************/
struct picture *
transform_get_picture(
	struct transform		*trf)

	{
	return (trf->pic);
	}



/****************************************************************************
 * transform_get_level
 ****************************************************************************/
int
transform_get_level(
	struct transform		*trf)

	{
	return (trf->level);
	}



/****************************************************************************
 * transform_get_wavelet
 ****************************************************************************/
int
transform_get_wavelet(
	struct transform		*trf)

	{
	return (trf->wavelet);
	}
/******************************************************** Karsten Scheibler */
