/****************************************************************************
 ****************************************************************************
 *
 * yawica.c
 *
 ****************************************************************************
 ****************************************************************************/





#include <stdio.h>
#include <string.h>

#include "error.h"
#include "debug.h"
#include "verbose.h"
#include "ac.h"
#include "codec.h"
#include "file.h"
#include "pgm.h"
#include "picture.h"
#include "psnr.h"
#include "transform.h"
#include "version.h"



#define PROGRAM_NAME			"yawica"
#define VERSION_STRING			PROGRAM_NAME " " VERSION "-" VERSION_DATE

#define YAWICA_MAX_FILES		2
#define YAWICA_MODE_VERSION		1
#define YAWICA_MODE_COMPARE		2
#define YAWICA_MODE_ENCODE		3
#define YAWICA_MODE_DECODE		4



char					program_name[] = PROGRAM_NAME;
static int				mode;
static char				*files[YAWICA_MAX_FILES];
static int				nfiles = 0;
static double				bpp;
static int				aclimit;
static int				bitlimit;
static int				bitplane;
static int				nr_bitplanes = 32;
static int				pass = 256;
static int				level;
static int				codec = 7;
static int				wavelet = TRANSFORM_WAVELET_53;



/****************************************************************************
 * set_limits
 ****************************************************************************/
static void
set_limits(
	int				xres,
	int				yres)

	{
	if (bpp > 0.0) aclimit = ((int) (bpp * (double) (xres * yres))) / 8;
	if (aclimit <= 0) return;
	aclimit -= file_get_header_size() + ac_get_flush_bytes();
	if (aclimit <= 0) error_message("file header larger than specified size");
	}



/****************************************************************************
 * cmd_version
 ****************************************************************************/
static void
cmd_version(
	void)

	{
	printf(VERSION_STRING "\n");
	}



/****************************************************************************
 * cmd_compare
 ****************************************************************************/
static void
cmd_compare(
	void)

	{
	struct picture			pic1, pic2;
	struct psnr			psn;

	pgm_read(files[0], &pic1);
	pgm_read(files[1], &pic2);
	psnr_calculate(&pic1, &pic2, &psn);
	psnr_print(&psn);
	}



/****************************************************************************
 * cmd_encode
 ****************************************************************************/
static void
cmd_encode(
	void)

	{
	struct picture			pic;
	struct transform		trf;
	struct ac			ac;
	struct codec			cdc;
	struct file			fil;
	int				xres, yres;

	pgm_read(files[0], &pic);
	xres  = picture_get_xres(&pic);
	yres  = picture_get_yres(&pic);
	level = transform_get_max_level(xres, yres, level);
	set_limits(xres, yres);
	transform_init(&trf, &pic, level, wavelet);
	transform_forward(&trf);
	transform_convert(&trf);

	/*
	 * reserve 2 * xres * yres bytes = 16 bpp
	 * for the compressed picture
	 */

	ac_init(&ac, NULL, 2 * xres * yres, bitlimit, aclimit, AC_MODE_WRITE, 0);
	codec_init(&cdc, &ac, &trf, codec, bitplane, nr_bitplanes, pass);
	codec_encode(&cdc);
	file_init(&fil, &trf, &cdc);
	file_write(&fil, files[1]);
	ac_deinit(&ac);
	picture_deinit(&pic);
	}



/****************************************************************************
 * cmd_decode
 ****************************************************************************/
static void
cmd_decode(
	void)

	{
	struct picture			pic;
	struct transform		trf;
	struct ac			ac;
	struct codec			cdc;
	struct file			fil;
	int				xres, yres;

	file_read(&fil, files[0]);
	xres = file_get_xres(&fil);
	yres = file_get_yres(&fil);
	set_limits(xres, yres);
	picture_init(&pic, xres, yres);
	picture_write_uchar(&pic, NULL);
	transform_init(&trf, &pic, file_get_level(&fil), file_get_wavelet(&fil));
	transform_convert(&trf);
	if ((bitlimit <= 0) || (bitlimit > file_get_bitlimit(&fil))) bitlimit = file_get_bitlimit(&fil);
	ac_init(&ac, file_get_data(&fil), file_get_size(&fil), bitlimit, aclimit, AC_MODE_READ, 0);
	codec_init(&cdc, &ac, &trf, file_get_codec(&fil), bitplane, nr_bitplanes, pass);
	codec_set_msb(&cdc, file_get_msb(&fil));
	codec_decode(&cdc);
	if ((ac_get_uncompressed_bits(&ac) < bitlimit) && ((aclimit > file_get_size(&fil)) ||
		(aclimit <= 0))) error_warning("file '%s' is truncated", files[0]);
	picture_set_shift(&pic, file_get_shift(&fil));
	transform_unconvert(&trf);
	transform_reverse(&trf);
	pgm_write(files[1], &pic);
	ac_deinit(&ac);
	picture_deinit(&pic);
	}



/****************************************************************************
 * cmdline_fill_space
 ****************************************************************************/
static void
cmdline_fill_space(
	char				*string)

	{
	while (*string != '\0') *string++ = ' ';
	}



/****************************************************************************
 * cmdline_usage
 ****************************************************************************/
static void
cmdline_usage(
	void)

	{
	char				space1[] = VERSION_STRING;
	char				space2[] = PROGRAM_NAME;

	cmdline_fill_space(space1);
	cmdline_fill_space(space2);
	printf(VERSION_STRING " | written by Karsten Scheibler\n"
		"%s | http://unusedino.de/yawica/\n"
		"%s | yawica@unusedino.de\n\n"
		"Usage: %s -V\n"
		"or:    %s -C [-v] [--] <file1> <file2>\n"
		"or:    %s -E [-v] [-p <float>] [-s <int>] [-b <int>] [-n <int>] [-c <int>]\n"
		"       %s    [-r <int>] [-l <int>] [-t <name>] [--] <srcfile> <dstfile>\n"
		"or:    %s -D [-v] [-p <float>] [-s <int>] [-b <int>] [-n <int>]\n"
		"       %s    [-r <int>] [--] <srcfile> <dstfile>\n\n"
		"  -V           print out version\n"
		"  -C           compare 2 pgm files and print out PSNR\n"
		"  -E           encode\n"
		"  -D           decode\n"
		"  -v           be more verbose\n"
#ifdef YAWICA_DEBUG
		"  -d           increase debug level\n"
#endif /* YAWICA_DEBUG */
		"  -p <float>   specify bits per pixel\n"
		"  -s <int>     specify file size in bytes\n"
		"  -b <int>     specify uncompressed bit count\n"
		"  -n <int>     specify bitplane\n"
		"  -r <int>     specify number of bitplanes\n"
		"  -c <int>     specify codec\n"
		"  -l <int>     specify number of transformation levels\n"
		"  -t <name>    select transformation by name\n",
		space1, space1, program_name, program_name, program_name,
		space2, program_name, space2);
	exit(0);
	}



/****************************************************************************
 * cmdline_parse
 ****************************************************************************/
static void
cmdline_parse(
	int				argc,
	char				**argv)

	{
	char				*arg;
	int				args, ignore = 0;

	for (args = 0, argv++; (arg = *argv++) != NULL; args++)
		{
		if ((arg[0] != '-') || (ignore))
			{
			if (nfiles >= YAWICA_MAX_FILES) error_message("too many files given");
			files[nfiles++] = arg;
			}
		else if (strcmp(arg, "--") == 0)
			{
			ignore = 1;
			}
		else if ((strcmp(arg, "-h") == 0) || (strcmp(arg, "--help") == 0))
			{
			cmdline_usage();
			}
		else if (((strcmp(arg, "-V") == 0) || (strcmp(arg, "--version") == 0)) && (args == 0))
			{
			mode = YAWICA_MODE_VERSION;
			}
		else if (((strcmp(arg, "-C") == 0) || (strcmp(arg, "--compare") == 0)) && (args == 0))
			{
			mode = YAWICA_MODE_COMPARE;
			}
		else if (((strcmp(arg, "-E") == 0) || (strcmp(arg, "--encode") == 0)) && (args == 0))
			{
			mode = YAWICA_MODE_ENCODE;
			}
		else if (((strcmp(arg, "-D") == 0) || (strcmp(arg, "--decode") == 0)) && (args == 0))
			{
			mode = YAWICA_MODE_DECODE;
			}
		else if ((mode == 0) || (mode == YAWICA_MODE_VERSION))
			{
			goto bad_option;
			}
		else if ((strcmp(arg, "-v") == 0) || (strcmp(arg, "--verbose") == 0))
			{
			verbose_level++;
			}
#ifdef YAWICA_DEBUG
		else if ((strcmp(arg, "-d") == 0) || (strcmp(arg, "--debug") == 0))
			{
			debug_level++;
			}
#endif /* YAWICA_DEBUG */
		else if (mode == YAWICA_MODE_COMPARE)
			{
			goto bad_option;
			}
		else if ((strcmp(arg, "-p") == 0) || (strcmp(arg, "--bpp") == 0))
			{
			int		i = 0;

			aclimit = 0;
			if (*argv != NULL) i = sscanf(*argv++, "%lf", &bpp);
			if ((i != 1) || (bpp <= 0.0)) error_message("-p/--bpp expects a positive floating point number");
			}
		else if ((strcmp(arg, "-s") == 0) || (strcmp(arg, "--file-size") == 0))
			{
			int		i = 0;

			bpp = 0.0;
			if (*argv != NULL) i = sscanf(*argv++, "%d", &aclimit);
			if ((i != 1) || (aclimit < 0)) error_message("-s/--file-size expects a positive decimal number");
			}
		else if ((strcmp(arg, "-b") == 0) || (strcmp(arg, "--bit-limit") == 0))
			{
			int		i = 0;

			if (*argv != NULL) i = sscanf(*argv++, "%d", &bitlimit);
			if ((i != 1) || (aclimit < 0)) error_message("-b/--bit-limit expects a positive decimal number");
			}
		else if ((strcmp(arg, "-n") == 0) || (strcmp(arg, "--bitplane") == 0))
			{
			int		i = 0, j = 2;

			if (*argv != NULL)
				{
				i = sscanf(*argv, "%d:%d", &bitplane, &pass);
				if (i != 2) j = 1, i = sscanf(*argv, "%d", &bitplane);
				argv++;
				}
			if ((i != j) || (bitplane < 0) || (bitplane > 31) || (pass < 0)) error_message("-n/--bitplane expects a positive decimal number");
			}
		else if ((strcmp(arg, "-r") == 0) || (strcmp(arg, "--nr-bitplanes") == 0))
			{
			int		i = 0;

			if (*argv != NULL) i = sscanf(*argv++, "%d", &nr_bitplanes);
			if ((i != 1) || (nr_bitplanes < 1) || (nr_bitplanes > 32)) error_message("-r/--nr-bitplanes expects a positive decimal number");
			}
		else if (mode != YAWICA_MODE_ENCODE)
			{
			goto bad_option;
			}
		else if ((strcmp(arg, "-c") == 0) || (strcmp(arg, "--codec") == 0))
			{
			int		i = 0;

			if (*argv != NULL) i = sscanf(*argv++, "%d", &codec);
			if ((i != 1) || (! codec_number_ok(codec))) error_message("-c/--codec expects a positive decimal number");
			}
		else if ((strcmp(arg, "-l") == 0) || (strcmp(arg, "--level") == 0))
			{
			int		i = 0;

			if (*argv != NULL) i = sscanf(*argv++, "%d", &level);
			if ((i != 1) || (level < 1)) error_message("-l/--level expects a positive decimal number");
			}
		else if ((strcmp(arg, "-t") == 0) || (strcmp(arg, "--transform") == 0))
			{
			wavelet = transform_get_wavelet_number(*argv++);
			if (wavelet == TRANSFORM_NONE) error_message("-t/--transform expects a name");
			}
		else
			{
		bad_option:
			error("unrecognized option '%s'", arg);
			}
		}
	if ((mode == 0) || ((mode != YAWICA_MODE_VERSION) && (nfiles < 2))) error_message("too few parameters given");
	}



/****************************************************************************
 * main
 ****************************************************************************/
int
main(
	int				argc,
	char				**argv)

	{

	/* parse cmdline */

	cmdline_parse(argc, argv);

	/* decide what to do */

	if (mode == YAWICA_MODE_VERSION) cmd_version();
	else if (mode == YAWICA_MODE_COMPARE) cmd_compare();
	else if (mode == YAWICA_MODE_ENCODE) cmd_encode();
	else if (mode == YAWICA_MODE_DECODE) cmd_decode();
	else debug_error();

	/* done */

	return (0);
	}
/******************************************************** Karsten Scheibler */
